<?php
namespace WCProductsDependencies;

/**
 * WCProductsDependencies Admin Class
 *
 * @class Admin
 * @version 2.0.2
 */
class Admin
{
    // <editor-fold desc="Properties">
    /**
     * Global settings model variable
     * @var array
     */
    public $globalSettingsModel = [];

    /**
     * WC products settings model variable
     * @var array
     */
    public $productSettingsModel = [];

    /**
     * WC product variations settings model variable
     * @var array
     */
    public $productVariationSettingsModel = [];

    /**
     * WC products categories settings model variable
     * @var array
     */
    public $productCategorySettingsModel = [];
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Class Constructor */
    public function __construct()
    {
        $settingsModel = Settings::getModel();

        $this->globalSettingsModel = $settingsModel['global'];
        $this->productSettingsModel = $settingsModel['product'];
        $this->productVariationSettingsModel = $settingsModel['productVariation'];
        $this->productCategorySettingsModel = $settingsModel['productCategory'];

        // admin scripts enqueue
        add_action('admin_enqueue_scripts', [$this, 'enqueueAssets'], 10);

        // core
        add_action('admin_init', [$this, 'initAction']);
        add_filter('admin_notices', [$this, 'noticesAction'], 10, 2);

        // product post
        add_action('add_meta_boxes', [$this, 'addMetaBoxes']);
        add_action('save_post_product', [$this, 'saveProductSettings']);

        // product category fields
        add_action('product_cat_edit_form_fields', [$this, 'outputProductCategoryFields'], 10);
        add_action('edited_product_cat', [$this, 'saveProductCategorySettings'], 10);

        // product variation fields
        add_action('woocommerce_product_after_variable_attributes', [$this, 'outputProductVariationFields'], 10, 3);
        add_action('woocommerce_process_product_meta_variable', [$this, 'saveProductVariationSettings'], 10);
        add_action('woocommerce_save_product_variation', [$this, 'saveProductVariationSettings'], 10);

        // settings page
        add_filter('woocommerce_settings_tabs_array', [$this, 'addSettingsTab'], 50);
        add_action('woocommerce_settings_tabs_products_dependencies_settings_tab', [$this, 'initSettingsTab']);
        add_action('woocommerce_update_options_products_dependencies_settings_tab', [$this, 'updateSettingsTab']);

        // orders
        add_filter('woocommerce_hidden_order_itemmeta', [$this, 'hiddenOrderMetaFilter']);

        // ajax
        add_action('wp_ajax_woocommerce_json_search_terms', [$this, 'jsonSearchTerms']);
    }

    /**
     * Generate html-field from passed args
     *
     * @param array $modelItem
     * @param array $args
     */
    public static function settingFieldView($modelItem, $args = array())
    {
        $defaultArgs = [
            'values' => [],
            'namePattern' => '%key%',
            'idPattern' => '%key%',
            'generateId' => true,
            'asTemplate' => false
        ];

        $args = array_replace($defaultArgs, $args);

        // create name from pattern
        $args['name'] = str_replace('%key%', $modelItem['key'], $args['namePattern']);

        // define value
        $args['value'] = isset($args['values'][$modelItem['key']])
            ? $args['values'][$modelItem['key']]
            : (isset($modelItem['default']) ? $modelItem['default'] : '');

        // define id attribute
        $args['id'] = $args['generateId'] ? str_replace('%key%', $modelItem['key'], $args['idPattern']) : false;

        $viewPath = __DIR__ . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR
            . 'setting-field-views' . DIRECTORY_SEPARATOR . $modelItem['type'] . '.php';

        echo isset($modelItem['before']) ? $modelItem['before'] : '';

        if (file_exists($viewPath)) {
            include($viewPath);
        }

        echo isset($modelItem['after']) ? $modelItem['after'] : '';

        if (isset($modelItem['description'])) {
            echo '<p class="description">' . wp_kses_post($modelItem['description']) . '</p>';
        }
    }

    /** Admin styles and scripts enqueue */
    public function enqueueAssets()
    {
        $path = WC_PRODUCTS_DEPENDENCIES_DEBUG ? 'src' : 'assets';
        $suffix = WC_PRODUCTS_DEPENDENCIES_DEBUG ? '' : '.min';
        $stylesFolder = WC_PRODUCTS_DEPENDENCIES_DEBUG ? 'scss' : 'css';

        wp_enqueue_script('jquery');
        wp_enqueue_script('jquery-ui-sortable');
        wp_enqueue_script(
            'wcpd-data-table',
            WC_PRODUCTS_DEPENDENCIES_PLUGIN_URL . "$path/admin/js/data-table$suffix.js",
            [
                'jquery',
                'jquery-ui-sortable'
            ],
            WC_PRODUCTS_DEPENDENCIES_VERSION,
            true
        );

        wp_enqueue_script(
            'wcpd-hooks',
            WC_PRODUCTS_DEPENDENCIES_PLUGIN_URL . "$path/admin/js/hooks$suffix.js",
            'jquery',
            WC_PRODUCTS_DEPENDENCIES_VERSION,
            true
        );

        wp_enqueue_style(
            'wcpd',
            WC_PRODUCTS_DEPENDENCIES_PLUGIN_URL . "$path/admin/$stylesFolder/app$suffix.css",
            [],
            WC_PRODUCTS_DEPENDENCIES_VERSION
        );
    }

    /** Admin notices hook */
    public function noticesAction()
    {
        require_once(__DIR__ . DIRECTORY_SEPARATOR . 'Legacy.php');

        $version = get_option('woocommerce_products_dependencies_version', null);

        if (is_null($version)) {
            $version = WC_PRODUCTS_DEPENDENCIES_VERSION;
            update_option('woocommerce_products_dependencies_version', WC_PRODUCTS_DEPENDENCIES_VERSION);
        }

        if (version_compare($version, Legacy::$lastChangedVersion, '<')) {
            $updateURL = wp_nonce_url(
                add_query_arg('do_update_woocommerce_products_dependencies', 'true'),
                'wcpd_db_update',
                'wcpd_db_update_nonce'
            );

            $dismissURL = wp_nonce_url(
                add_query_arg('dismiss_update_woocommerce_products_dependencies', 'true'),
                'wcpd_db_update_dismiss',
                'wcpd_db_update_dismiss_nonce'
            );
            ?>
            <div class="notice notice-warning is-dismissible">
                <p><strong><?php
                        esc_html_e(
                            'WooCommerce Products Dependencies database update required',
                            'woocommerce-products-dependencies'
                        );
                        ?></strong></p>
                <p><?php
                    esc_html_e(
                        'WooCommerce Products Dependencies has been updated! To keep things running smoothly, we have to update your database to the newest version.',
                        'woocommerce-products-dependencies'
                    );
                    ?></p>
                <p>
                    <a href="<?php echo esc_url($updateURL); ?>" class="wcpd-update-now button-primary"><?php
                        esc_html_e(
                            'Update WooCommerce Products Dependencies Database',
                            'woocommerce-products-dependencies'
                        );
                        ?></a>
                    <a href="<?php echo esc_url($dismissURL); ?>" class="wcpd-update-dismiss button"><?php
                        esc_html_e('Dismiss', 'woocommerce-products-dependencies');
                        ?></a>
                </p>
            </div>
            <?php
        }
    }

    /** Admin init hook */
    public function initAction()
    {
        if (!empty($_GET['dismiss_update_woocommerce_products_dependencies'])) { // WPCS: input var ok.
            check_admin_referer('wcpd_db_update_dismiss', 'wcpd_db_update_dismiss_nonce');
            update_option('woocommerce_products_dependencies_version', WC_PRODUCTS_DEPENDENCIES_VERSION);
        }

        if (!empty($_GET['init_update_woocommerce_products_dependencies'])) { // WPCS: input var ok.
            check_admin_referer('wcpd_db_update_init', 'wcpd_db_update_init_nonce');

            require_once(__DIR__ . DIRECTORY_SEPARATOR . 'Legacy.php');

            // init Legacy class
            new Legacy();

            update_option('woocommerce_products_dependencies_version', WC_PRODUCTS_DEPENDENCIES_VERSION);

            add_action('admin_notices', function () {
                // phpcs:disable
                echo '<div class="notice notice-success is-dismissible"><p>'
                    . esc_html__(
                        'WooCommerce Products Dependencies database update complete. Thank you for updating to the latest version!',
                        'woocommerce-products-dependencies'
                    )
                    . '</p></div>';
                // phpcs:enable
            });
        }
    }

    /** Add posts meta-boxes */
    public function addMetaBoxes()
    {
        add_meta_box(
            'woocommerce-products-dependencies',
            esc_html__('Products Dependencies', 'woocommerce-products-dependencies'),
            [$this, 'outputProductFields'],
            'product',
            'normal'
        );
    }
    // </editor-fold>

    // <editor-fold desc="Product">
    /**
     * Output product meta fields
     *
     * @param \WP_Post $post
     */
    public function outputProductFields($post)
    {
        ?>
        <table class="wcpd wcpd-settings-table form-table">
            <?php foreach ($this->productSettingsModel as $key => $setting) { ?>
                <tr valign="top">
                    <th scope="row" class="titledesc">
                        <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                            echo wp_kses_post($setting['label']);
                            ?></label>
                    </th>
                    <td><?php
                        self::settingFieldView(
                            $setting,
                            ['values' => [$setting['key'] => Settings::getProduct($post->ID, $key)]]
                        );
                        ?></td>
                </tr>
            <?php } ?>
        </table>
        <?php
    }

    /**
     * Save product meta fields
     *
     * @param integer $postId
     */
    public function saveProductSettings($postId)
    {
        if ((defined('DOING_AUTOSAVE') && DOING_AUTOSAVE)
            || !current_user_can('edit_post', $postId)
            || (isset($_REQUEST['action']) && $_REQUEST['action'] == 'inline-save')
        ) {
            return;
        }

        foreach ($this->productSettingsModel as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_post_meta($postId, $setting['key'], $_POST[$setting['key']]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Product category">
    /**
     * Output product category meta fields
     *
     * @param \WP_Term $term
     */
    public function outputProductCategoryFields($term)
    {
        foreach ($this->productCategorySettingsModel as $key => $setting) {
            ?>
            <tr class="wcpd wcpd-settings-table" valign="top">
                <th scope="row" class="titledesc">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                        echo wp_kses_post($setting['label']);
                        ?></label>
                </th>
                <td><?php
                    self::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => Settings::getProductCategory($term->term_id, $key)]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save product category meta fields
     *
     * @param integer $termId
     */
    public function saveProductCategorySettings($termId)
    {
        if (!$termId) {
            return;
        }

        foreach ($this->productCategorySettingsModel as $setting) {
            if (!isset($_POST[$setting['key']])) {
                continue;
            }

            update_term_meta($termId, $setting['key'], $_POST[$setting['key']]);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Product variation">
    /**
     * Output product variation meta fields
     *
     * @param integer $loop
     * @param array   $variationData
     * @param \WP_Post  $variation
     */
    public function outputProductVariationFields($loop, $variationData, $variation)
    {
        foreach ($this->productVariationSettingsModel as $key => $setting) {
            ?>
            <div class="wcpd wcpd-settings-table">
                <p>
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                        echo wp_kses_post($setting['label']);
                        ?></label>
                </p>
                <?php
                self::settingFieldView(
                    $setting,
                    [
                        'namePattern' => "%key%[$loop]",
                        'idPattern' => "%key%-$loop",
                        'values' => [$setting['key'] => Settings::getProductVariation($variation->ID, $key)]
                    ]
                );
                ?>
            </div>
            <?php
        }
    }

    /** Save product variation meta values */
    public function saveProductVariationSettings()
    {
        if (!isset($_POST['variable_post_id']) || !$_POST['variable_post_id']) {
            return;
        }

        foreach ($_POST['variable_post_id'] as $index => $variationId) {
            foreach ($this->productVariationSettingsModel as $setting) {
                if (!isset($_POST[$setting['key']][$index])) {
                    continue;
                }

                update_post_meta($variationId, $setting['key'], $_POST[$setting['key']][$index]);
            }
        }
    }
    // </editor-fold>

    // <editor-fold desc="Order">
    /**
     * Order product hidden meta filter
     *
     * @param array $items
     *
     * @return array
     */
    public function hiddenOrderMetaFilter($items)
    {
        $items[] = '_wcpd_prepayment_bundle_size';

        return $items;
    }
    // </editor-fold>

    // <editor-fold desc="Settings tab">
    /**
     * Add the tab to the woocommerce options page
     *
     * @param array $settingsTabs
     *
     * @return array $settingsTabs
     */
    public function addSettingsTab($settingsTabs)
    {
        $settingsTabs['products_dependencies_settings_tab'] = 'Products Dependencies';

        return $settingsTabs;
    }

    /** Init options fields */
    public function initSettingsTab()
    {
        $defaults = [
            'section_title' => [
                'name' => 'Products Dependencies',
                'type' => 'title',
                'desc' => '',
                'id' => 'wc_products_dependencies_section_title'
            ]
        ];

        $args = array_merge($defaults, $this->globalSettingsModel);

        woocommerce_admin_fields($args);
        woocommerce_admin_fields(
            [
                'section_end' => [
                    'type' => 'sectionend',
                    'id' => 'wc_products_dependencies_section_end'
                ]
            ]
        );
    }

    /** Save options fields */
    public function updateSettingsTab()
    {
        woocommerce_update_options($this->globalSettingsModel);
    }
    // </editor-fold>

    // <editor-fold desc="Utils">
    /**Search for terms and echo json */
    public function jsonSearchTerms()
    {
        ob_start();
        check_ajax_referer('search-products', 'security');

        $args = [
            'taxonomy' => 'product_cat',
            'orderby' => 'id',
            'order' => 'ASC',
            'hide_empty' => false,
            'include' => !empty($_GET['include']) ? array_map('intval', explode(',', $_GET['include'])) : '',
            'exclude' => !empty($_GET['exclude']) ? array_map('intval', explode(',', $_GET['exclude'])) : '',
            'number' => !empty($_GET['limit']) ? intval($_GET['limit']) : '',
            'name__like' => $_GET['term']
        ];

        $terms = get_terms($args);
        $foundTerms = [];

        if (!empty($terms)) {
            foreach ($terms as $term) {
                $foundTerms[$term->term_id] = rawurldecode($term->name);
            }
        }

        $foundTerms = apply_filters('woocommerce_json_search_found_terms', $foundTerms);

        wp_send_json($foundTerms);
    }
    // </editor-fold>
}
