<?php
namespace WCProductsDependencies;

/**
 * Core Class
 *
 * @class Core
 * @version 6.2.0
 */
class Core
{
    // <editor-fold desc="Properties">
    /**
     * Self instance variable
     * @var Core The single instance of the class
     */
    protected static $instance = null;

    /**
     * Admin class instance
     * @var Admin
     */
    public static $admin;

    /**
     * Front class instance
     * @var Front
     */
    public static $front;
    // </editor-fold>

    //<editor-fold desc="Core">
    /** Core Constructor */
    public function __construct()
    {
        // include base slave classes
        $requiredClasses = [
            'Settings',
            'Front'
        ];

        foreach ($requiredClasses as $requiredClass) {
            if (!class_exists('\\WCProductsDependencies\\' . $requiredClass)
                && file_exists(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php')
            ) {
                include_once(__DIR__ . DIRECTORY_SEPARATOR . $requiredClass . '.php');
            }
        }

        self::$front = new Front();

        // actions
        add_action('plugins_loaded', [$this, 'loadTextDomain']);
        add_action('woocommerce_init', [$this, 'onWoocommerceInit'], 1);

        // orders hooks
        add_action('woocommerce_payment_complete', [$this, 'orderCompletedAction']);
        add_action('woocommerce_order_status_completed', [$this, 'orderCompletedAction']);
    }

    /**
     * Get single class instance
     *
     * @static
     * @return Core
     */
    public static function instance()
    {
        if (self::$instance === null) {
            self::$instance = new self();
        }

        return self::$instance;
    }

    /** Load text domain */
    public function loadTextDomain()
    {
        load_plugin_textdomain(
            'woocommerce-products-dependencies',
            false,
            basename(WC_PRODUCTS_DEPENDENCIES_PLUGIN_PATH) . '/languages/'
        );
    }

    /** Handles then woocommerce plugin is loaded */
    public function onWoocommerceInit()
    {
        if (is_admin()) {
            if (!class_exists('\\WCProductsDependencies\\Admin')) {
                include_once(__DIR__ . DIRECTORY_SEPARATOR . 'Admin.php');
            }

            self::$admin = new Admin();
        }
    }
    //</editor-fold>

    //<editor-fold desc="Order">
    /**
     * Get user prepaid products meta
     *
     * @param integer $userId
     *
     * @return array
     */
    public static function getUserPrepaidProducts($userId = null)
    {
        $userId = $userId ? $userId : get_current_user_id();

        return (array) get_user_meta($userId, '_wcpd_prepaid_products', 1);
    }

    /**
     * Set user prepaid products meta
     *
     * @param array $value
     * @param integer $userId
     *
     * @return bool
     */
    public static function setUserPrepaidProducts($value, $userId = null)
    {
        $userId = $userId ? $userId : get_current_user_id();
        $value = array_filter($value, function ($item) {
            return $item !== '';
        });

        return update_user_meta($userId, '_wcpd_prepaid_products', $value);
    }

    /**
     * Or order/payment complete
     *
     * @param integer $orderId
     */
    public function orderCompletedAction($orderId)
    {
        // if order is already handled
        if (get_post_meta($orderId, '_wcpd_prepayment_bundle_handled', 1)) {
            return;
        }

        $order = wc_get_order($orderId);
        $items = $order->get_items();
        $userId = $order->get_customer_id();
        $userPrepaidProducts = self::getUserPrepaidProducts($userId);

        foreach ($items as $item) {
            $quantity = $item->get_quantity();
            $meta = $item->get_meta_data();
            $product = $item->get_product();
            $productId = $product->get_id();

            foreach ($meta as $metaItem) {
                $itemData = $metaItem->get_data();

                if ($itemData['key'] != '_wcpd_prepayment_bundle_size') {
                    continue;
                }

                if (!isset($userPrepaidProducts[$productId])) {
                    $userPrepaidProducts[$productId] = (int) $itemData['value'] - $quantity;
                } else {
                    $userPrepaidProducts[$productId] -= $quantity;
                }

                $userPrepaidProducts[$productId] = max(0, $userPrepaidProducts[$productId]);

                break;
            }
        }

        update_post_meta($orderId, '_wcpd_prepayment_bundle_handled', 'yes');
        self::setUserPrepaidProducts($userPrepaidProducts, $userId);
    }
    //</editor-fold>

    //<editor-fold desc="Utils">
    /**
     * Get product's terms array
     *
     * @param integer $productId
     *
     * @return array
     */
    public static function getProductTerms($productId)
    {
        static $wcpdProductTermsCache = [];

        if (!isset($wcpdProductTermsCache[$productId])) {
            $wcpdProductTermsCache[$productId] = wp_get_post_terms($productId, 'product_cat');
        }

        return $wcpdProductTermsCache[$productId];
    }

    /**
     * Get product's terms ids array
     *
     * @param integer $productId
     *
     * @return array
     */
    public static function getProductTermsIds($productId)
    {
        static $wcpdProductTermsIdsCache = [];

        if (!isset($wcpdProductTermsIdsCache[$productId])) {
            $wcpdProductTermsIdsCache[$productId] = wp_get_post_terms($productId, 'product_cat', ['fields' => 'ids']);
        }

        return $wcpdProductTermsIdsCache[$productId];
    }

    /**
     * Prepare dependencies array
     *
     * @param array $input
     *
     * @return array $input
     */
    public static function prepareDependenciesArray($input)
    {
        foreach ($input as $key => &$value) {
            if (is_string($value)) {
                $value = array_filter(explode(',', $value));
            }

            // if there is no dependencies
            if (empty($value) || count($value) != 4 || !(!empty($value['terms']) || !empty($value['products']))) {
                unset($input[$key]);
            }
        }

        return array_filter($input);
    }

    /**
     * Get products ids from the each customer's order
     *
     * @param integer $userId
     *
     * @return array
     */
    public static function getUserOrderedProducts($userId = null)
    {
        $output = [];

        if (!$userId) {
            $userId = get_current_user_id();
        }

        if (!$userId) {
            return $output;
        }

        // get all customer orders
        $orders = wc_get_orders(apply_filters(
            'wcpd_user_orders_query_args',
            [
                'numberposts' => -1,
                'status' => 'completed',
                'customer_id' => $userId
            ]
        ));

        // get products ids from the each customer's order
        foreach ($orders as $order) {
            $orderItems = $order->get_items();

            foreach ($orderItems as $orderItem) {
                if (isset($output[$orderItem['product_id']])) {
                    $output[$orderItem['product_id']] += $orderItem['quantity'];
                } else {
                    $output[$orderItem['product_id']] = $orderItem['quantity'];
                }
            }
        }

        return $output;
    }

    /**
     * Get product minimum quantity value from the option
     *
     * @param integer $id
     *
     * @return integer
     */
    public static function getProductMinQuantityValue($id)
    {
        $valuesToCheck = [];
        $output = [1];

        if (Settings::getProduct($id, 'enable_individual_min_quantity')) {
            // check by product individually
            $valuesToCheck[] = Settings::getProduct($id, 'min_quantity');
        } else {
            // check by product terms
            foreach (self::getProductTermsIds($id) as $termId) {
                $valuesToCheck[] = Settings::getProductCategory($termId, 'min_product_quantity');
            }
        }

        // check values
        foreach ($valuesToCheck as $value) {
            $results = self::checkProductCountSettingValue($value);

            if (!$results['hasRule']) {
                continue;
            }

            $output[] = $results['value'];
        }

        return max($output);
    }

    /**
     * Check product count (quantity and selected) setting value
     *
     * @param integer|array $value - simple count (older versions) or terms rules
     * @param string $type - "quantity" or "selected"
     *
     * @return array
     */
    public static function checkProductCountSettingValue($value, $type = 'quantity')
    {
        $cartCounts = [];

        /**
         * older versions support
         * @since 1.13.0
         */
        if (is_integer($value)) {
            $value = [
                'type' => 'count',
                'count' => $value,
                'terms' => []
            ];
        }

        // value isn't defined
        if (!isset($value['type']) || ($value['type'] == 'count' && !$value['count'])
            || ($value['type'] != 'count' && empty($value['terms']))
        ) {
            return [
                'hasRule' => 0,
                'value' => null
            ];
        }

        // check selected count
        if ($value['type'] == 'terms-count' && $type == 'selected') {
            $cartCounts = self::getCartCounts(self::getCartContent(), $value['terms']);

            return [
                'hasRule' => 1,
                'value' => count($cartCounts['products'])
            ];
        }

        // check quantities count
        if ($value['type'] != 'count') {
            $cartCounts = self::getCartCounts(self::getCartContent(), $value['terms']);
        }

        switch ($value['type']) {
            case 'terms-least':
                $count = min($cartCounts['products']);
                break;

            case 'terms-greatest':
                $count = max($cartCounts['products']);
                break;

            case 'terms-sum':
                $count = array_sum($cartCounts['products']);
                break;

            default:
            case 'count':
                $count = (int) $value['count'];
        }

        return [
            'hasRule' => 1,
            'value' => $count
        ];
    }
    //</editor-fold>

    //<editor-fold desc="Cart">
    /**
     * Collect cart (and old orders if needed) items ids, terms and quantities
     *
     * @param array $cart
     * @param array $termsIds - specific terms only
     *
     * @return array
     */
    public static function getCartCounts($cart = null, $termsIds = null)
    {
        if (is_null($cart)) {
            $cart = self::getCartContent();
        }

        $output = [
            'products' => [],
            'terms' => []
        ];

        // get user orders products
        if (filter_var(Settings::getGlobal('check_customer_ordered_products'), FILTER_VALIDATE_BOOLEAN)) {
            $output['products'] = self::getUserOrderedProducts();
        }

        // get all cart products ids and quantities
        foreach ($cart as $cartItem) {
            if (isset($output['products'][$cartItem['product_id']])) {
                // merge with the previous values
                $output['products'][$cartItem['product_id']] += $cartItem['quantity'];
            } else {
                $output['products'][$cartItem['product_id']] = $cartItem['quantity'];
            }

            if ($cartItem['variation_id']) {
                if (isset($output['products'][$cartItem['variation_id']])) {
                    // merge with the previous values
                    $output['products'][$cartItem['variation_id']] += $cartItem['quantity'];
                } else {
                    $output['products'][$cartItem['variation_id']] = $cartItem['quantity'];
                }
            }
        }

        // get all cart terms ids and quantities
        foreach ($output['products'] as $id => $quantity) {
            $productTerms = self::getProductTermsIds($id);

            // keep specific terms only
            if (!empty($termsIds) && empty(array_intersect($productTerms, $termsIds))) {
                unset($output['products'][$id]);
            }

            // calculate products quantities
            foreach ($productTerms as $termId) {
                // keep specific terms only
                if (!empty($termsIds) && !in_array($termId, $termsIds)) {
                    continue;
                }

                if (isset($output['terms'][$termId])) {
                    $output['terms'][$termId] += $quantity;
                } else {
                    $output['terms'][$termId] = $quantity;
                }
            }
        }

        return apply_filters('wcpd_cart_quantities', $output, $cart);
    }

    /**
     * Return cart content from the woocommerce storage
     *
     * @return array
     */
    public static function getCartContent()
    {
        return apply_filters('wcpd_cart_content', \WC()->cart->get_cart());
    }
    //</editor-fold>

    //<editor-fold desc="Dependencies">
    /**
     * Get products ids which are superior for the product
     *
     * @param integer $productId
     * @param integer $variationId
     * @param array $productTerms
     *
     * @return array
     */
    public static function getProductSuperiorsIds($productId, $variationId, $productTerms = [])
    {
        $termsDependencies = Settings::getProduct($productId, 'product_categories_dependencies');
        $variationTermsDependencies = Settings::getProductVariation($variationId, 'product_categories_dependencies');
        $termsDependencies = array_merge($termsDependencies, $variationTermsDependencies);
        $productsDependencies = Settings::getProduct($productId, 'products_dependencies');
        $variationDependencies = Settings::getProductVariation($variationId, 'products_dependencies');

        if (empty($productTerms)) {
            $productTerms = self::getProductTermsIds($productId);
        }

        // if have no meta
        if (empty($productsDependencies) || !is_array($productsDependencies)) {
            $productsDependencies = [];
        }

        // if have no meta
        if (empty($variationDependencies) || !is_array($variationDependencies)) {
            $variationDependencies = [];
        }

        // get terms dependencies
        foreach ($productTerms as $productTerm) {
            $termDependency = Settings::getProductCategory($productTerm, 'product_categories_dependencies');

            if (empty($termDependency) || !is_array($termDependency)) {
                continue;
            }

            $termsDependencies = array_merge($termsDependencies, $termDependency);
        }

        // prepare and clean arrays
        $termsDependencies = self::prepareDependenciesArray($termsDependencies);
        $productsDependencies = self::prepareDependenciesArray($productsDependencies);
        $variationDependencies = self::prepareDependenciesArray($variationDependencies);

        $output = [
            'terms' => $termsDependencies,
            'products' => $productsDependencies,
            'variations' => $variationDependencies
        ];

        return apply_filters('wcpd_product_superiors_ids', $output, $productId, $variationId, $productTerms);
    }

    /**
     * Check product quantity dependency
     *
     * @param integer $productId
     * @param integer $variationId
     *
     * @return bool
     */
    public static function hasQuantityDependency($productId, $variationId = 0)
    {
        $productDependency = Settings::getProduct($productId, 'quantity_dependency');
        $productTerms = self::getProductTermsIds($productId);
        $termDependency = false;
        $variationDependency = false;

        if ($variationId) {
            $variationDependency = Settings::getProductVariation($variationId, 'quantity_dependency');
        }

        foreach ($productTerms as $productTerm) {
            $termDependency = $termDependency || Settings::getProductCategory($productTerm, 'quantity_dependency');
        }

        if (!$variationDependency && !$productDependency && !$termDependency) {
            return false;
        }

        return true;
    }

    /**
     * Get product quantity dependency ratio setting value
     *
     * @param integer $productId
     * @param integer $variationId
     *
     * @return bool
     */
    public static function getQuantityDependencyRatio($productId, $variationId = 0)
    {
        // check variation
        if ($variationId) {
            $variationRatio = Settings::getProductVariation($variationId, 'quantity_dependency_ratio');

            if ($variationRatio) {
                return $variationRatio;
            }
        }

        $productRatio = Settings::getProduct($productId, 'quantity_dependency_ratio');

        // check product
        if ($productRatio) {
            return $productRatio;
        }

        // check terms
        foreach (self::getProductTermsIds($productId) as $productTerm) {
            $termRatio = Settings::getProductCategory($productTerm, 'quantity_dependency_ratio');

            if ($termRatio) {
                return $termRatio;
            }
        }

        return 1;
    }

    /**
     * Check product can be purchased for by no-logged user
     *
     * @param integer $productId
     * @param integer $variationId
     *
     * @return bool
     */
    public static function getProductLoggedDependency($productId, $variationId = 0)
    {
        // already logged
        if (is_user_logged_in()) {
            return true;
        }

        // check options
        if (Settings::getGlobal('logged_dependency') || Settings::getProduct($productId, 'logged_dependency')
            || ($variationId && Settings::getProductVariation($variationId, 'logged_dependency'))
        ) {
            return false;
        }

        // get terms dependencies
        foreach (self::getProductTermsIds($productId) as $productTerm) {
            if (Settings::getProductCategory($productTerm, 'logged_dependency')) {
                return false;
            }
        }

        return true;
    }

    /**
     * Check product can be purchased more than once
     *
     * @param integer $productId
     *
     * @return bool
     */
    public static function checkBuyOnceDependency($productId)
    {
        // already logged
        if (!is_user_logged_in()) {
            return true;
        }

        $checkOrders = false;

        if (Settings::getProduct($productId, 'buy_once')) {
            $checkOrders = true;
        } else {
            foreach (self::getProductTermsIds($productId) as $productTerm) {
                if (Settings::getProductCategory($productTerm, 'buy_once')) {
                    $checkOrders = true;

                    break;
                }
            }
        }

        if ($checkOrders) {
            $ordered = self::getUserOrderedProducts();

            if (isset($ordered[$productId])) {
                return false;
            }
        }

        return true;
    }

    /**
     * Check product can be in the cart
     *
     * @param array $args
     *
     * @return array
     */
    public static function checkProductDependencies($args = [])
    {
        $defaults = [
            'productId' => null,
            'variationId' => null,
            'quantity' => 1,
            'cartCounts' => null,
            'advanceMode' => false
        ];

        $args = array_replace($defaults, $args);
        $output = [];

        if (is_null($args['cartCounts'])) {
            $args['cartCounts'] = self::getCartCounts();
        }

        $productsInCart = $args['cartCounts']['products'];
        $termsInCart = $args['cartCounts']['terms'];
        $quantityInTheCart = isset($productsInCart[$args['productId']]) ? $productsInCart[$args['productId']] : 0;
        $productTerms = self::getProductTermsIds($args['productId']);
        $hasQuantityDependency = self::hasQuantityDependency($args['productId'], $args['variationId']);
        $quantityDependencyRatio = self::getQuantityDependencyRatio($args['productId'], $args['variationId']);
        $superiorsIds = self::getProductSuperiorsIds($args['productId'], $args['variationId'], $productTerms);
        $superiorsProducts = array_merge($superiorsIds['products'], $superiorsIds['variations']);
        $termsMetWeight = 0;
        $productsMetWeight = 0;
        $isTermsQuantityMet = empty($superiorsIds['terms']);
        $isProductsQuantityMet = empty($superiorsProducts);
        $previousWeight = null;
        $previousIsRowMinQuantityMet = null;
        $productQuantityToCompare = $args['quantity'];
        $termQuantityToCompare = $args['quantity'];

        if ($args['variationId'] && isset($superiorsIds['products']) && isset($productsInCart[$args['productId']])) {
            // get quantity of all variations for a variable product
            $productQuantityToCompare = $productsInCart[$args['productId']];
        }

        // check by logged dependency
        if (!self::getProductLoggedDependency($args['productId'], $args['variationId'])) {
            $output['logged'] = true;
        }

        // check buy-once dependency
        if (!self::checkBuyOnceDependency($args['productId'])) {
            $output['buy-once'] = true;

            return apply_filters('wcpd_product_dependencies', $output, $args);
        }

        // check products incompatibility
        if ($args['variationId'] && !self::checkProductsIncompatibility($args['productId'], $args['variationId'])) {
            $variationLevel = (array) Settings::getProductVariation($args['variationId'], 'incompatible_products');
            $productLevel = (array) Settings::getProduct($args['productId'], 'incompatible_products');
            $incompatibleProducts = array_unique(array_merge($variationLevel, $productLevel));
            $output['products-incompatibility'] = $incompatibleProducts;
        } elseif (!self::checkProductsIncompatibility($args['productId'])) {
            $incompatibleProducts = (array) Settings::getProduct($args['productId'], 'incompatible_products');
            $output['products-incompatibility'] = $incompatibleProducts;
        }

        foreach ($productTerms as $productTerm) {
            // check terms incompatibility
            if (!self::checkTermsIncompatibility($productTerm, array_keys($termsInCart))) {
                $incompatibleTerms = (array) Settings::getProductCategory($productTerm, 'incompatible_product_categories');
                $output['terms-incompatibility'] = $incompatibleTerms;
            }

            // get the term quantity to compare with the quantity rules further
            if (isset($termsInCart[$productTerm])) {
                $termQuantityToCompare = $termsInCart[$productTerm];
            }
        }

        // increment quantity for advance mode
        if ($args['advanceMode']) {
            $productQuantityToCompare += $quantityInTheCart;
            $termQuantityToCompare += 1;
        }

        // if have no meta values
        if (empty($superiorsProducts)) {
            $productsMetWeight = 1;
        }

        if (empty($superiorsIds['terms'])) {
            $termsMetWeight = 1;
        }

        // check by terms
        foreach ($superiorsIds['terms'] as $dependencyRow) {
            if (!isset($dependencyRow['terms']) || empty($dependencyRow['terms'])) {
                continue;
            }

            $rowWeight = 0;
            $difference = array_diff($dependencyRow['terms'], array_keys($termsInCart));
            $intersect = array_intersect_key($termsInCart, array_flip($dependencyRow['terms']));

            // get the row value
            if ($dependencyRow['inner_relation'] == 'and' && empty($difference)) {
                $rowWeight = min($intersect);
            } elseif ($dependencyRow['inner_relation'] == 'or' && !empty($intersect)) {
                $rowWeight = max($intersect);
            }

            $isRowMinQuantityMet = $rowWeight >= (int) $dependencyRow['min_quantity'];

            // look at the previous values
            if (!is_null($previousWeight)) {
                $rowWeight = min($rowWeight, $previousWeight);
            }

            if (!is_null($previousIsRowMinQuantityMet)) {
                $isRowMinQuantityMet = min($previousIsRowMinQuantityMet, $isRowMinQuantityMet);
            }

            if ($dependencyRow['outer_relation'] == 'and' && $dependencyRow != end($superiorsIds['terms'])) {
                $previousWeight = $rowWeight;
                $previousIsRowMinQuantityMet = $isRowMinQuantityMet;
                $isRowMinQuantityMet = false;
            } else {
                $previousWeight = null;
                $previousIsRowMinQuantityMet = null;
                $termsMetWeight = max($termsMetWeight, $rowWeight);
            }

            // check product quantity
            if (($termsMetWeight && $termsMetWeight * $quantityDependencyRatio >= $termQuantityToCompare || !$hasQuantityDependency)
                && $isRowMinQuantityMet
            ) {
                $isTermsQuantityMet = true;
            }
        }

        // check by products
        foreach ($superiorsProducts as $dependencyRow) {
            if (!isset($dependencyRow['products']) || empty($dependencyRow['products'])) {
                continue;
            }

            $rowWeight = 0;
            $difference = array_diff($dependencyRow['products'], array_keys($productsInCart));
            $intersect = array_intersect_key($productsInCart, array_flip($dependencyRow['products']));

            // get the row value
            if ($dependencyRow['inner_relation'] == 'and' && empty($difference)) {
                $rowWeight = min($intersect);
            } elseif ($dependencyRow['inner_relation'] == 'or' && !empty($intersect)) {
                $rowWeight = max($intersect);
            }

            $isRowMinQuantityMet = $rowWeight >= (int) $dependencyRow['min_quantity'];

            // look at the previous values
            if (!is_null($previousWeight)) {
                $rowWeight = min($rowWeight, $previousWeight);
            }

            if (!is_null($previousIsRowMinQuantityMet)) {
                $isRowMinQuantityMet = min($previousIsRowMinQuantityMet, $isRowMinQuantityMet);
            }

            if ($dependencyRow['outer_relation'] == 'and' && $dependencyRow != end($superiorsProducts)) {
                $previousWeight = $rowWeight;
                $previousIsRowMinQuantityMet = $isRowMinQuantityMet;
                $isRowMinQuantityMet = false;
            } else {
                $previousWeight = null;
                $previousIsRowMinQuantityMet = null;
                $productsMetWeight = max($productsMetWeight, $rowWeight);
            }

            // check product quantity
            if (($productsMetWeight && $productsMetWeight * $quantityDependencyRatio >= $productQuantityToCompare || !$hasQuantityDependency)
                && $isRowMinQuantityMet
            ) {
                $isProductsQuantityMet = true;
            }
        }

        if (!$termsMetWeight) {
            $output['terms'] = $superiorsIds['terms'];
        }

        if ($termsMetWeight && !$isTermsQuantityMet) {
            $output['terms-quantity'] = $superiorsIds['terms'];
        }

        if (!$productsMetWeight) {
            $output['products'] = $superiorsProducts;
        }

        if ($productsMetWeight && !$isProductsQuantityMet) {
            $output['products-quantity'] = $superiorsProducts;
        }

        return apply_filters('wcpd_product_dependencies', $output, $args);
    }

    /**
     * Get products min/max quantity rules errors
     *
     * @return array
     */
    public static function getProductsQuantityRulesErrors()
    {
        $cart = self::getCartContent();
        $cartTerms = [];
        $output = [
            'min_products_quantity' => [],
            'min_products_selected' => []
        ];

        foreach ($cart as $cartItem) {
            $productTerms = self::getProductTerms($cartItem['product_id']);

            // get product terms total values
            foreach ($productTerms as $productTerm) {
                if (!isset($cartTerms[$productTerm->term_id])) {
                    $cartTerms[$productTerm->term_id] = $productTerm;
                    $cartTerms[$productTerm->term_id]->productsPerTerm = 0;
                    $cartTerms[$productTerm->term_id]->productsSelected = 0;
                }

                $cartTerms[$productTerm->term_id]->productsPerTerm += (int) $cartItem['quantity'];
                $cartTerms[$productTerm->term_id]->productsSelected += 1;
            }
        }

        foreach ($cartTerms as $cartTerm) {
            $minProductsQuantity = Settings::getProductCategory($cartTerm->term_id, 'min_products_quantity');
            $minProductsQuantityValue = self::checkProductCountSettingValue($minProductsQuantity);
            $minProductsSelected = Settings::getProductCategory($cartTerm->term_id, 'min_products_selected');
            $minProductsSelectedValue = self::checkProductCountSettingValue($minProductsSelected, 'selected');

            if ($minProductsQuantityValue['hasRule']
                && $cartTerm->productsPerTerm < $minProductsQuantityValue['value']
            ) {
                $output['min_products_quantity'][] = [
                    'term' => $cartTerm,
                    'value' => $minProductsQuantityValue['value']
                ];
            }

            if ($minProductsSelectedValue['hasRule']
                && $cartTerm->productsSelected < $minProductsSelectedValue['value']
            ) {
                $output['min_products_selected'][] = [
                    'term' => $cartTerm,
                    'value' => $minProductsSelectedValue['value']
                ];
            }
        }

        return apply_filters('wcpd_product_quantity_rules_errors', $output);
    }

    /**
     * Get dependencies errors
     *
     * @return array
     */
    public static function getDependenciesErrors()
    {
        $cart = self::getCartContent();
        $cartCounts = self::getCartCounts($cart);
        $output = [];

        foreach ($cart as $cartItem) {
            $dependencies = self::checkProductDependencies([
                'productId' => $cartItem['product_id'],
                'variationId' => $cartItem['variation_id'],
                'quantity' => $cartItem['quantity'],
                'cartCounts' => $cartCounts
            ]);

            if (!empty($dependencies)) {
                $output[] = self::$front->getDependencyMessage($dependencies, $cartItem['product_id'], $cartItem['variation_id']);
            }
        }

        return apply_filters('wcpd_cart_dependency_errors', $output);
    }

    /**
     * Check terms incompatibility with products in the cart
     *
     * @param integer $termId
     * @param array $termsInCart
     *
     * @return bool
     */
    public static function checkTermsIncompatibility($termId, $termsInCart)
    {
        $incompatibleTerms = (array) Settings::getProductCategory($termId, 'incompatible_product_categories');

        if (empty($incompatibleTerms)) {
            return true;
        }

        if (!empty(array_intersect($termsInCart, $incompatibleTerms))) {
            return false;
        }

        return true;
    }

    /**
     * Check product incompatibility with products in the cart
     *
     * @param integer $productId
     * @param integer $variationId
     *
     * @return bool
     */
    public static function checkProductsIncompatibility($productId, $variationId = 0)
    {
        $incompatibleProducts = (array) Settings::getProduct($productId, 'incompatible_products');

        if ($variationId) {
            $incompatibleProducts = array_merge(
                $incompatibleProducts,
                (array) Settings::getProductVariation($variationId, 'incompatible_products')
            );
        }

        $incompatibleProducts = array_map('intval', $incompatibleProducts);

        if (empty($incompatibleProducts)) {
            return true;
        }

        $cart = self::getCartContent();
        $cartProductsIds = [];

        // get all cart products ids and quantities
        foreach ($cart as $cartItem) {
            $cartProductsIds[] = $cartItem['product_id'];
            $cartProductsIds[] = $cartItem['variation_id'];
        }

        if (!empty(array_intersect($cartProductsIds, $incompatibleProducts))) {
            return false;
        }

        return true;
    }
    //</editor-fold>
}
