<?php
namespace WCProductsDependencies;

/**
 * Front Class
 *
 * @class Front
 * @version 4.1.0
 */
class Front
{
    // <editor-fold desc="Properties">
    /**
     * Messages shown session key variable
     * @var string
     */
    public $messageShowedSessionKey = 'wcpd_message_showed';
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Front Constructor */
    public function __construct()
    {
        // start session if it doesn't exist
        if (!session_id()) {
            session_start();
        }

        // product
        add_filter('woocommerce_product_is_visible', [$this, 'productVisibilityFilter'], 10, 2);
        add_action('woocommerce_simple_auctions_before_place_bid_filter', [$this, 'placeSimpleAuctionBidFilter'], 10);

        // cart hooks
        add_action('woocommerce_update_cart_action_cart_updated', [$this, 'handleCartDependencies'], 10, 1);
        add_action('woocommerce_before_calculate_totals', [$this, 'cartTotalsAction'], 25);

        // cart item add to cart
        add_action('woocommerce_add_to_cart', [$this, 'addToCartAction'], 15, 6);
        add_filter('woocommerce_add_cart_item', [$this, 'addToCartItemFilter'], 10, 1);
        add_filter('woocommerce_add_cart_item_data', [$this, 'addToCartItemDataFilter'], 20, 3);

        // cart item remove
        add_action('woocommerce_cart_item_removed', [$this, 'handleCartDependencies'], 10, 1);

        // cart item quantity update
        add_action('woocommerce_after_cart_item_quantity_update', [$this, 'handleCartDependencies'], 10, 1);

        // cart item entities
        add_filter('woocommerce_cart_item_price', [$this, 'cartItemPriceFilter'], 10, 2);
        add_filter('woocommerce_cart_item_name', [$this, 'cartItemNameFilter'], 20, 3);
        add_filter('woocommerce_quantity_input_args', [$this, 'quantityInputFilter'], 10, 2);

        // orders hooks
        add_action('woocommerce_checkout_create_order', [$this, 'orderCreateAction'], 20);
        add_action('woocommerce_checkout_create_order_line_item', [$this, 'orderProductDataAction'], 20, 4);

        // pages hooks
        add_action('woocommerce_check_cart_items', [$this, 'checkoutPageAction']);
        add_action('woocommerce_before_main_content', [$this, 'singleProductAction'], 10);
    }
    // </editor-fold>

    // <editor-fold desc="Order">
    /**
     * Order data action
     *
     * @param \WC_Order $order
     *
     * @throws \Exception
     */
    public function orderCreateAction($order)
    {
        $max_orders_per_period = Settings::getGlobal('max_orders_per_period');
        $max_orders_time_period = Settings::getGlobal('max_orders_time_period');

        if ($max_orders_per_period && $max_orders_time_period) {
            $currentTimestamp = time();
            $max_orders_time_periodLimit = round($max_orders_time_period * 30);
            $userOrders = get_posts([
                'numberposts' => $max_orders_per_period,
                'post_type' => 'shop_order',
                'post_status' => array_keys(wc_get_order_statuses()),
                'orderby' => 'date',
                'order' => 'DESC',
                'date_query' => [
                    [
                        'inclusive' => true,
                        'before' => date('Y/m/d H:i:s', $currentTimestamp + $max_orders_time_periodLimit)
                    ],
                    [
                        'inclusive' => true,
                        'after' => date('Y/m/d H:i:s', $currentTimestamp - $max_orders_time_periodLimit)
                    ]
                ],
                'meta_query' => [
                    [
                        'meta_key' => '_customer_user',
                        'meta_value' => get_current_user_id()
                    ],
                    [
                        'meta_key' => '_customer_ip_address',
                        'meta_value' => $order->get_customer_ip_address()
                    ]
                ]
            ]);

            if (count($userOrders) >= $max_orders_per_period) {
                throw new \Exception(Settings::getGlobal('max_orders_per_period_message'));
            }
        }
    }

    /**
     * Order product data action
     *
     * @param \WC_Order_Item_Product $item
     * @param string $cartItemKey
     * @param array $values
     * @param \WC_Order $order
     */
    public function orderProductDataAction($item, $cartItemKey, $values, $order)
    {
        if (isset($values['wcpd_prepayment_bundle_size']) && $values['wcpd_prepayment_bundle_size']) {
            $item->add_meta_data('_wcpd_prepayment_bundle_size', $values['wcpd_prepayment_bundle_size']);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Cart item">
    /**
     * Cart item data filter on adding
     *
     * @param array $args
     *
     * @return array
     */
    public function addToCartItemFilter($args)
    {
        // filter product quantity before adding to the cart
        $args['quantity'] = max($this->getProductMinQuantity($args['product_id']), $args['quantity']);

        return $args;
    }

    /**
     * Product data filter on adding
     *
     * @param array $cartItemData
     * @param integer $productId
     * @param integer $variationId
     *
     * @return array
     */
    public function addToCartItemDataFilter($cartItemData, $productId, $variationId)
    {
        // save default product price for the prepaid bundles work only for logged in users
        if (!is_user_logged_in()) {
            return $cartItemData;
        }

        $id = $variationId > 0 ? $variationId : $productId;
        $productTermsIds = Core::getProductTermsIds($productId);

        foreach ($productTermsIds as $productTermId) {
            if (($bundleSize = Settings::getProductCategory($productTermId, 'prepayment_bundle_size')) && $bundleSize) {
                $bundleRegularity = Settings::getProductCategory($productTermId, 'prepayment_bundle_regularity');
                $product = wc_get_product($id);

                $cartItemData['wcpd_default_price'] = (float) $product->get_price();
                $cartItemData['wcpd_prepayment_bundle_size'] = (int) $bundleSize;
                $cartItemData['wcpd_prepayment_bundle_regularity'] = $bundleRegularity;

                return $cartItemData;
            }
        }

        return $cartItemData;
    }

    /**
     * Cart item price filter
     *
     * @param string $productPrice
     * @param array $cartItem
     *
     * @return string
     */
    public function cartItemPriceFilter($productPrice, $cartItem)
    {
        // replace changed price by default string for prepaid bundles
        if (isset($cartItem['wcpd_default_price'])) {
            $productPrice = wc_price(
                wc_get_price_to_display($cartItem['data'], ['price' => $cartItem['wcpd_default_price']])
            );
        }

        return $productPrice;
    }

    /**
     * Cart item name filter
     *
     * @param string $name
     * @param array $cartItem
     * @param string $cartItemKey
     *
     * @return string
     */
    public function cartItemNameFilter($name, $cartItem, $cartItemKey)
    {
        if (!is_user_logged_in() || !isset($cartItem['wcpd_default_price'])) {
            return $name;
        }

        // modify product title if it is in a prepayment bundle
        $quantity = $cartItem['quantity'];
        $bundleSize = $cartItem['wcpd_prepayment_bundle_size'];
        $productId = $cartItem['variation_id'] ? $cartItem['variation_id'] : $cartItem['product_id'];
        $bundleRegularity = $cartItem['wcpd_prepayment_bundle_regularity'];
        $userPrepaidProducts = Core::getUserPrepaidProducts();
        $paid = isset($userPrepaidProducts[$productId]) ? $userPrepaidProducts[$productId] : 0;
        $message = '';

        // the bundle was paid already
        if (($bundleSize == $quantity || $bundleRegularity == 'first-order' && isset($userPrepaidProducts[$productId]))
            && !$paid
        ) {
            return $name;
        }

        // calculate payable items
        if ($bundleRegularity == 'each-order'
            || ($bundleRegularity == 'first-order' && $quantity + $paid < $bundleSize)
        ) {
            // for each order or the first order with not enough quantity
            $payable = $bundleSize * max(1, ceil(($quantity - $paid) / $bundleSize));
        } else {
            // for the first order with enough quantity
            $payable = $quantity - $paid;
        }

        if (!$paid) {
            // have no paid items
            if ($quantity < $bundleSize && $bundleRegularity == 'first-order') {
                // quantity less than first order's bundle size
                $message = Settings::getGlobal('prepayment_bundle_label');
            } elseif ($quantity % $bundleSize != 0 && $bundleRegularity == 'each-order') {
                // quantity isn't aliquots to bundle size for each order
                $message = Settings::getGlobal('prepayment_bundle_label');
            }
        } elseif ($quantity > $paid && $bundleRegularity == 'each-order') {
            // have not enough paid items and need to pay for the new bundle
            $message = Settings::getGlobal('prepayment_bundle_with_paid_products_label');
        } else {
            // have enough paid items
            $message = Settings::getGlobal('paid_products_label');
        }

        $replacements = apply_filters(
            'wcpd_prepayment_bundle_message_entities',
            [
                '$bundle$' => $bundleSize,
                '$payable$' => $payable,
                '$paid$' => $paid
            ],
            $name,
            $cartItem,
            $cartItemKey
        );

        $message = str_replace(array_keys($replacements), array_values($replacements), $message);
        $message = " <span class=\"wcpd-prepayment-bundle\">$message</span>";
        $name .= apply_filters('wcpd_prepayment_bundle_message', $message, $name, $cartItem, $cartItemKey);

        return $name;
    }
    // </editor-fold>

    // <editor-fold desc="Quantity rules">
    /**
     * Prepare message of term minimum products error
     *
     * @param string $type
     * @param \WP_Term $term
     * @param string $count
     *
     * @return string
     */
    public function getProductsQuantityErrorMessage($type, $term, $count)
    {
        $message = esc_html__('Unknown dependency', 'woocommerce-products-dependencies');
        $messageKeys = [
            'min_products_quantity' => 'min_products_quantity_message',
            'min_products_selected' => 'min_products_selected_message'
        ];

        if (isset($messageKeys[$type])) {
            $message = Settings::getGlobal($messageKeys[$type]);
        }

        $replacements = apply_filters(
            'wcpd_products_quantity_error_message_entities',
            [
                '$n$' => $count,
                '$category$' => sprintf(
                    '<a href="%1$s" class="wcpd-message-item-link wcpd-%2$s">%3$s</a>',
                    get_term_link($term, 'product_cat'),
                    'term',
                    $term->name
                )
            ],
            $term,
            $count
        );

        $message = str_replace(array_keys($replacements), array_values($replacements), $message);

        return apply_filters('wcpd_products_quantity_error_message', $message);
    }

    /**
     * Get minimum quantity product based on the cart
     *
     * @param integer $productId
     *
     * @return integer
     */
    public function getProductMinQuantity($productId)
    {
        $cart = Core::getCartContent();
        $minQuantity = Core::getProductMinQuantityValue($productId);

        // no minimum options
        if (!$minQuantity) {
            return 1;
        }

        // is cart page
        if (is_cart()) {
            return $minQuantity;
        }

        // reduce minimum quantity if is in the cart
        foreach ($cart as $cartItem) {
            if ($cartItem['product_id'] == $productId) {
                $minQuantity -= (int) $cartItem['quantity'];

                break;
            }
        }

        $minQuantity = max($minQuantity, 1);

        return apply_filters('wcpd_product_min_quantity', $minQuantity, $productId);
    }

    /**
     * Product quantity input args filter
     *
     * @param array $args
     * @param object $product
     *
     * @return array
     */
    public function quantityInputFilter($args, $product)
    {
        $id = null;

        if ($product && method_exists($product, 'get_id')) {
            $id = $product->get_id();
        } elseif (isset($product->id) && $product->id) {
            $id = $product->id;
        }

        if (!$id) {
            return $args;
        }

        // set product's quantity input minimum value
        $value = $this->getProductMinQuantity($id);

        $args['input_value'] = max($value, $args['input_value']);
        $args['min_value'] = $value;

        return $args;
    }
    // </editor-fold>

    // <editor-fold desc="Dependencies">
    /**
     * Transform passed arguments to the message item string
     *
     * @param array $args
     *
     * @return string
     */
    public function getDependencyMessageItem($args)
    {
        $defaults = [
            'type' => 'none',
            'text' => '',
            'tag' => 'li'
        ];

        $args = array_merge($defaults, $args);

        switch ($args['type']) {
            case 'and':
                $args['text'] = Settings::getGlobal('dependency_message_and_string');
                break;

            case 'or':
                $args['text'] = Settings::getGlobal('dependency_message_or_string');
                break;

            case 'logged':
                $args['text'] = Settings::getGlobal('logged_dependency_message');
                break;

            case 'buy-once':
                $args['text'] = Settings::getGlobal('buy_once_dependency_message');
                break;

            case 'products-incompatibility':
                $args['text'] = Settings::getGlobal('products_incompatibility_message');
                break;

            case 'terms-incompatibility':
                $args['text'] = Settings::getGlobal('product_categories_incompatibility_message');
                break;

            case 'terms':
                $args['text'] = Settings::getGlobal('product_categories_dependency_message');
                break;

            case 'terms-quantity':
                $args['text'] = Settings::getGlobal('product_categories_quantity_dependency_message');
                break;

            case 'products':
                $args['text'] = Settings::getGlobal('products_dependency_message');
                break;

            case 'products-quantity':
                $args['text'] = Settings::getGlobal('products_quantity_dependency_message');
                break;
        }

        return apply_filters(
            'wcpd_message_item',
            sprintf(
                '<%1$s class="wcpd-message-item wcpd-%2$s">%3$s</%1$s>',
                $args['tag'],
                $args['type'],
                do_shortcode($args['text'])
            )
        );
    }

    /**
     * Compile product dependencies to the message string
     *
     * @param array $dependencies
     * @param integer $productId
     * @param integer $variationId
     *
     * @return string
     */
    public function getDependencyMessage($dependencies, $productId = null, $variationId = null)
    {
        $messages = [];
        $customMessage = null;
        $productTermsIds = Core::getProductTermsIds($productId);

        if ($variationId && Settings::getProductVariation($variationId, 'enable_custom_dependencies_message')) {
            // variation custom message
            $customMessage = Settings::getProductVariation($variationId, 'custom_dependencies_message');
        } elseif (Settings::getProduct($productId, 'enable_custom_dependencies_message')) {
            // product custom message
            $customMessage = Settings::getProduct($productId, 'custom_dependencies_message');
        } else {
            // categories custom message
            foreach ($productTermsIds as $productTermsId) {
                if (Settings::getProductCategory($productTermsId, 'enable_custom_dependencies_message')) {
                    $customMessage = Settings::getProductCategory($productTermsId, 'custom_dependencies_message');

                    break;
                }
            }
        }

        if ($customMessage) {
            $customMessage = '<div class="wcpd-message">' . do_shortcode($customMessage) . '</div>';

            return apply_filters('wcpd_message', $customMessage, $productId);
        }

        if (empty($dependencies)) {
            return esc_html__('Unknown dependency', 'woocommerce-products-dependencies');
        }

        foreach ($dependencies as $dependencyKey => $dependency) {
            $messages[] = $this->getDependencyMessageItem([
                'type' => $dependencyKey,
                'tag' => 'div'
            ]);

            switch ($dependencyKey) {
                case 'products':
                case 'products-quantity':
                    foreach ($dependency as $dependencyRow) {
                        $titles = [];

                        if (reset($dependency) == $dependencyRow) {
                            $messages[] = '<ul class="wcpd-message-item">';
                        }

                        foreach ($dependencyRow['products'] as $dependencyRowId) {
                            $link = get_permalink($dependencyRowId);
                            $title = get_the_title($dependencyRowId);

                            // if name or link are empty for some reasons
                            if (empty($title) || empty($link)) {
                                $product = get_posts([
                                    'post_type' => ['product', 'product_variation'],
                                    'include' => $dependencyRowId
                                ]);

                                $product = $product[0];
                                $link = $product->guid;
                                $title = $product->post_title;
                            }

                            $titles[] = apply_filters(
                                'wcpd_message_item_product',
                                sprintf(
                                    '<a href="%1$s" '
                                    . 'class="wcpd-message-item-link wcpd-%2$s">%3$s</a>',
                                    $link,
                                    'products',
                                    $title
                                ),
                                $dependencyRowId
                            );
                        }

                        $text = implode(
                            ' '
                            . $this->getDependencyMessageItem([
                                'type' => $dependencyRow['inner_relation'],
                                'tag' => 'span'
                            ])
                            . ' ',
                            $titles
                        );

                        if ($dependencyRow['min_quantity'] && $dependencyRow['min_quantity'] > 1) {
                            $text .= ' ('
                                . sprintf(
                                    esc_html__('minimum quantity: %1$d', 'woocommerce-products-dependencies'),
                                    $dependencyRow['min_quantity']
                                )
                                . ')';
                        }

                        $messages[] = $this->getDependencyMessageItem(['type' => 'product', 'text' => $text]);

                        if (end($dependency) == $dependencyRow) {
                            $messages[] = '</ul>';
                        } else {
                            $messages[] = $this->getDependencyMessageItem(['type' => $dependencyRow['outer_relation']]);
                        }
                    }

                    break;
                case 'terms':
                case 'terms-quantity':
                    foreach ($dependency as $dependencyRow) {
                        $titles = [];

                        if (reset($dependency) == $dependencyRow) {
                            $messages[] = '<ul class="wcpd-message-item">';
                        }

                        foreach ($dependencyRow['terms'] as $dependencyRowId) {
                            $term = get_term($dependencyRowId);

                            if (!is_object($term)) {
                                continue;
                            }

                            $titles[] = apply_filters(
                                'wcpd_message_item_term',
                                sprintf(
                                    '<a href="%1$s" '
                                    . 'class="wcpd-message-item-link wcpd-%2$s">%3$s</a>',
                                    get_term_link($term, 'product_cat'),
                                    'term',
                                    $term->name
                                ),
                                $term
                            );
                        }

                        $text = implode(
                            ' '
                            . $this->getDependencyMessageItem([
                                'type' => $dependencyRow['inner_relation'],
                                'tag' => 'span'
                            ])
                            . ' ',
                            $titles
                        );

                        if ($dependencyRow['min_quantity'] && $dependencyRow['min_quantity'] > 1) {
                            $text .= ' ('
                                . sprintf(
                                    esc_html__('minimum quantity: %1$d', 'woocommerce-products-dependencies'),
                                    $dependencyRow['min_quantity']
                                )
                                . ')';
                        }

                        $messages[] = $this->getDependencyMessageItem(['type' => 'term', 'text' => $text]);

                        if (end($dependency) == $dependencyRow) {
                            $messages[] = '</ul>';
                        } else {
                            $messages[] = $this->getDependencyMessageItem(['type' => $dependencyRow['outer_relation']]);
                        }
                    }

                    break;
                case 'products-incompatibility':
                    foreach ($dependency as $dependencyItem) {
                        $titles = [];

                        if (reset($dependency) == $dependencyItem) {
                            $messages[] = '<ul class="wcpd-message-item">';
                        }

                        $link = get_permalink($dependencyItem);
                        $title = get_the_title($dependencyItem);

                        // if name or link are empty for some reasons
                        if (empty($title) || empty($link)) {
                            $product = get_posts([
                                'post_type' => ['product', 'product_variation'],
                                'include' => $dependencyItem
                            ]);

                            $product = $product[0];
                            $link = $product->guid;
                            $title = $product->post_title;
                        }

                        $titles[] = apply_filters(
                            'wcpd_message_item_product',
                            sprintf(
                                '<a href="%1$s" class="wcpd-message-item-link wcpd-%2$s">%3$s</a>',
                                $link,
                                'products',
                                $title
                            ),
                            $dependencyItem
                        );

                        $text = implode(
                            ' '
                            . $this->getDependencyMessageItem(['type' => 'and', 'tag' => 'span'])
                            . ' ',
                            $titles
                        );

                        $messages[] = $this->getDependencyMessageItem(['type' => 'product', 'text' => $text]);

                        if (end($dependency) == $dependencyItem) {
                            $messages[] = '</ul>';
                        }
                    }

                    break;
                case 'terms-incompatibility':
                    foreach ($dependency as $dependencyRow) {
                        $titles = [];

                        if (reset($dependency) == $dependencyRow) {
                            $messages[] = '<ul class="wcpd-message-item">';
                        }

                        $term = get_term($dependencyRow);

                        if (is_object($term)) {
                            $titles[] = apply_filters(
                                'wcpd_message_item_term',
                                sprintf(
                                    '<a href="%1$s" '
                                    . 'class="wcpd-message-item-link wcpd-%2$s">%3$s</a>',
                                    get_term_link($term, 'product_cat'),
                                    'term',
                                    $term->name
                                ),
                                $term
                            );
                        }

                        $text = implode(
                            ' '
                            . $this->getDependencyMessageItem(['type' => 'and', 'tag' => 'span'])
                            . ' ',
                            $titles
                        );

                        $messages[] = $this->getDependencyMessageItem(['type' => 'term', 'text' => $text]);

                        if (end($dependency) == $dependencyRow) {
                            $messages[] = '</ul>';
                        }
                    }
            }
        }

        foreach ($messages as &$message) {
            $replacements = apply_filters(
                'wcpd_dependency_message_entities',
                ['$product$' => get_the_title($productId)],
                $dependencies,
                $productId,
                $variationId
            );

            $message = str_replace(array_keys($replacements), array_values($replacements), $message);
        }

        return implode('', apply_filters('wcpd_message', $messages, $productId));
    }

    /**
     * Check and remove depends products from the cart
     *
     * @param string $itemKey - cart item key
     */
    public function handleCartDependencies($itemKey = null)
    {
        if (Settings::getGlobal('check_dependencies_on_checkout_only')) {
            return;
        }

        $cart = Core::getCartContent();

        if (is_string($itemKey)) {
            unset($cart[$itemKey]);
        }

        $cartCounts = Core::getCartCounts($cart);

        // remove unmet products
        foreach ($cart as $cartItemKey => $cartItem) {
            $dependencies = Core::checkProductDependencies([
                'productId' => $cartItem['product_id'],
                'variationId' => $cartItem['variation_id'],
                'quantity' => $cartItem['quantity'],
                'cartCounts' => $cartCounts
            ]);

            if (!empty($dependencies)) {
                \WC()->cart->remove_cart_item($cartItemKey);
            }
        }
    }

    /**
     * Filter product visibility
     *
     * @param bool $isVisible
     * @param integer $productId
     *
     * @return bool
     */
    public function productVisibilityFilter($isVisible, $productId)
    {
        $visibilityByUsers = (array) Settings::getProduct($productId, 'visibility_by_users');

        if (!empty($visibilityByUsers) && !in_array(get_current_user_id(), $visibilityByUsers)) {
            return false;
        }

        return $isVisible;
    }
    // </editor-fold>

    // <editor-fold desc="Pages">
    /** Before cart total calculation action */
    public function cartTotalsAction()
    {
        if (is_admin() && !defined('DOING_AJAX') || !is_user_logged_in()
            || !apply_filters('wcpd_cart_totals_action', true)
        ) {
            return;
        }

        $userPrepaidProducts = Core::getUserPrepaidProducts();

        foreach (Core::getCartContent() as $cartItem) {
            // modify products prices in the cart for prepaid products
            if (!isset($cartItem['wcpd_default_price']) || !isset($cartItem['wcpd_prepayment_bundle_size'])) {
                continue;
            }

            $bundleSize = $cartItem['wcpd_prepayment_bundle_size'];
            $bundleRegularity = $cartItem['wcpd_prepayment_bundle_regularity'];
            $productId = $cartItem['variation_id'] ? $cartItem['variation_id'] : $cartItem['product_id'];
            $quantity = $cartItem['quantity'];
            $paid = isset($userPrepaidProducts[$productId]) ? $userPrepaidProducts[$productId] : 0;
            $newPrice = 0;

            // the bundle was paid already
            if ($bundleRegularity == 'first-order' && isset($userPrepaidProducts[$productId]) && !$paid) {
                continue;
            }

            // calculate payable items
            if ($bundleRegularity == 'each-order'
                || ($bundleRegularity == 'first-order' && $quantity + $paid < $bundleSize)
            ) {
                // for each order or the first order with not enough quantity
                $payable = $bundleSize * max(1, ceil(($quantity - $paid) / $bundleSize));
            } else {
                // for the first order with enough quantity
                $payable = $quantity - $paid;
            }

            if ($quantity > $paid) {
                $newPrice = $cartItem['data']->get_price() * $payable / $quantity;
            }

            $newPrice = apply_filters('wcpd_prepayment_bundle_product_price', $newPrice, $cartItem, $userPrepaidProducts);

            $cartItem['data']->set_price($newPrice);
        }
    }

    /**
     * Add product to cart action
     *
     * @param string  $cartItemKey
     * @param integer $productId
     * @param integer $quantity
     * @param integer $variationId
     *
     * @throws \Exception
     */
    public function addToCartAction($cartItemKey, $productId, $quantity, $variationId)
    {
        if (Settings::getGlobal('check_dependencies_on_checkout_only')) {
            return;
        }

        $cart = Core::getCartContent();
        $dependencies = Core::checkProductDependencies([
            'productId' => $productId,
            'variationId' => $variationId,
            'quantity' => isset($cart[$cartItemKey]) ? $cart[$cartItemKey]['quantity'] : $quantity,
            'cartCounts' => Core::getCartCounts($cart)
        ]);

        // check product can be purchasable
        if (!empty($dependencies)) {
            // reduce product quantity as it already added
            $reducedQuantity = \WC()->cart->cart_contents[$cartItemKey]['quantity'] - $quantity;

            \WC()->cart->set_quantity($cartItemKey, $reducedQuantity, false);

            // save info about showing message
            $_SESSION[$this->messageShowedSessionKey] = true;

            throw new \Exception($this->getDependencyMessage($dependencies, $productId, $variationId));
        }
    }

    /** Single product page action */
    public function singleProductAction()
    {
        if (Settings::getGlobal('check_dependencies_on_checkout_only')) {
            return;
        }

        if (!is_single()
            || !Settings::getGlobal('output_dependency_messages_in_advance')
            || (isset($_SESSION[$this->messageShowedSessionKey]) && $_SESSION[$this->messageShowedSessionKey])
        ) {
            // reset storage value
            $_SESSION[$this->messageShowedSessionKey] = false;

            return;
        }

        $productId = get_the_ID();
        $quantityErrors = Core::getProductsQuantityRulesErrors();
        $dependencies = Core::checkProductDependencies([
            'productId' => $productId,
            'advanceMode' => true
        ]);

        if (!empty($dependencies)) {
            wc_add_notice($this->getDependencyMessage($dependencies, $productId), 'notice');
        }

        foreach ($quantityErrors as $type => $values) {
            if (!empty($values)) {
                foreach ($values as $value) {
                    wc_add_notice(
                        $this->getProductsQuantityErrorMessage($type, $value['term'], $value['value']),
                        'notice'
                    );
                }
            }
        }
    }

    /**
     * Add simple auction bid hook
     *
     * @param object $productData
     *
     * @return bool
     */
    public function placeSimpleAuctionBidFilter($productData)
    {
        if (Settings::getGlobal('check_dependencies_on_checkout_only')) {
            return true;
        }

        $dependencies = Core::checkProductDependencies(['productId' => $productData->get_id()]);

        if (!empty($dependencies)) {
            $_SESSION[$this->messageShowedSessionKey] = true;

            wc_add_notice($this->getDependencyMessage($dependencies, $productData->get_id()), 'error');

            return false;
        }

        return true;
    }

    /**
     * Check cart dependencies/rules on the checkout page and revert to cart in case of issues
     *
     * @return bool
     */
    public function checkoutPageAction()
    {
        // check terms products quantity rules
        if (!is_checkout()) {
            return true;
        }

        $quantityErrors = Core::getProductsQuantityRulesErrors();
        $dependencyErrors = Core::getDependenciesErrors();
        $hasErrors = false;

        // check dependencies
        if (!empty($dependencyErrors)) {
            $hasErrors = true;

            foreach ($dependencyErrors as $error) {
                wc_add_notice($error, 'error');
            }
        }

        // check quantity rules
        foreach ($quantityErrors as $type => $values) {
            if (!empty($values)) {
                $hasErrors = true;

                foreach ($values as $value) {
                    wc_add_notice(
                        $this->getProductsQuantityErrorMessage($type, $value['term'], $value['value']),
                        'error'
                    );
                }
            }
        }

        if (!$hasErrors) {
            return true;
        }

        wp_redirect(wc_get_page_permalink('cart'));

        exit;
    }
    // </editor-fold>
}
